# -*- coding: utf-8 -*-
from __future__ import absolute_import
import zlib
from .compression import CompressionSocket


class DefalteSocket(CompressionSocket):
    def __init__(self, socket, level):
        wbits = -zlib.MAX_WBITS
        self._decompressor = zlib.decompressobj(wbits)
        self._compressor = zlib.compressobj(level, zlib.DEFLATED, wbits)
        super(DefalteSocket, self).__init__(socket)

    def compress(self, data):
        data = self._compressor.compress(data)
        return data + self._compressor.flush(zlib.Z_SYNC_FLUSH)

    def decompress(self, data):
        return self._decompressor.decompress(data)

    def close(self):
        self._socket.sendall(self._compressor.flush(zlib.Z_FINISH))
        self._socket.close()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from .consumer import Consumer
from .reader import Reader
from .producer import Producer
from .nsqd import Nsqd, NsqdTCPClient, NsqdHTTPClient
from .lookupd import Lookupd, LookupdClient
from .message import Message
from .backofftimer import BackoffTimer
from .version import __version__

__author__ = 'Trevor Olson'
__email__ = 'trevor@heytrevor.com'
__version__ = __version__

__all__ = [
    'Consumer',
    'Reader',
    'Producer',
    'Nsqd',
    'NsqdTCPClient',
    'NsqdHTTPClient',
    'Lookupd',
    'LookupdClient',
    'Message',
    'BackoffTimer',
]



from __future__ import with_statement

import pytest
import gevent
import gnsq

from integration_server import (
    LookupdIntegrationServer,
    NsqdIntegrationServer
)


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_basic():
    with LookupdIntegrationServer() as server:
        lookupd = gnsq.LookupdClient(server.address, server.http_port)
        assert lookupd.ping() == b'OK'
        assert 'version' in lookupd.info()

        with pytest.raises(gnsq.errors.NSQHttpError):
            lookupd.lookup('topic')

        assert len(lookupd.topics()['topics']) == 0
        assert len(lookupd.channels('topic')['channels']) == 0
        assert len(lookupd.nodes()['producers']) == 0


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_lookup():
    with LookupdIntegrationServer() as lookupd_server:
        nsqd_server = NsqdIntegrationServer(lookupd=lookupd_server.tcp_address)
        with NsqdIntegrationServer(lookupd=lookupd_server.tcp_address) as nsqd_server:
            lookupd = gnsq.LookupdClient(lookupd_server.address, lookupd_server.http_port)
            nsqd = gnsq.NsqdHTTPClient(nsqd_server.address, nsqd_server.http_port)
            gevent.sleep(0.1)

            assert len(lookupd.topics()['topics']) == 0
            assert len(lookupd.channels('topic')['channels']) == 0
            assert len(lookupd.nodes()['producers']) == 1

            nsqd.create_topic('topic')
            gevent.sleep(0.1)

            info = lookupd.lookup('topic')
            assert len(info['channels']) == 0
            assert len(info['producers']) == 1
            assert len(lookupd.topics()['topics']) == 1
            assert len(lookupd.channels('topic')['channels']) == 0

            nsqd.create_channel('topic', 'channel')
            gevent.sleep(0.1)

            info = lookupd.lookup('topic')
            assert len(info['channels']) == 1
            assert len(info['producers']) == 1
            assert len(lookupd.topics()['topics']) == 1
            assert len(lookupd.channels('topic')['channels']) == 1



# -*- coding: utf-8 -*-
import gevent.queue


class QueueHandler(gevent.queue.Queue):
    """Iterator like api for gnsq.

    Example usage::

        >>> queue = QueueHandler()
        >>> consumer = Consumer('topic', 'worker', max_in_flight=16)
        >>> consumer.on_message.connect(queue)
        >>> consumer.start(block=False)
        >>> for message in queue:
        ...     print(message.body)
        ...     message.finish()

    Or give it to a pool::

        >>> gevent.pool.Pool().map(queue, my_handler)

    :param maxsize: maximum number of messages that can be queued. If less than
        or equal to zero or None, the queue size is infinite.
    """

    def __call__(self, consumer, message):
        message.enable_async()
        self.put(message)


class ChannelHandler(gevent.queue.Channel):
    """Iterator like api for gnsq.

    Like :class:`QueueHandler` with a ``maxsize`` of ``1``.
    """

    def __call__(self, consumer, message):
        message.enable_async()
        self.put(message)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division

import logging
import random
import time

from collections import defaultdict
from itertools import cycle

import blinker
import gevent

from gevent.event import Event
from gevent.pool import Group

from .backofftimer import BackoffTimer
from .decorators import cached_property
from .errors import NSQException, NSQRequeueMessage, NSQSocketError
from .nsqd import NsqdTCPClient
from .states import INIT, RUNNING, BACKOFF, THROTTLED, CLOSED
from .util import parse_nsqds, parse_lookupds


class Consumer(object):
    """High level NSQ consumer.

    A Consumer will connect to the nsqd tcp addresses or poll the provided
    nsqlookupd http addresses for the configured topic and send signals to
    message handlers connected to the :attr:`on_message` signal or provided by
    ``message_handler``.

    Messages will automatically be finished when the message handle returns
    unless :meth:`message.enable_async() <gnsq.Message.enable_async>` is called.
    If an exception occurs or :class:`~gnsq.errors.NSQRequeueMessage` is raised,
    the message will be requeued.

    The Consumer will handle backing off of failed messages up to a configurable
    ``max_interval`` as well as automatically reconnecting to dropped
    connections.

    Example usage::

        from gnsq import Consumer

        consumer = gnsq.Consumer('topic', 'channel', 'localhost:4150')

        @consumer.on_message.connect
        def handler(consumer, message):
            print 'got message:', message.body

        consumer.start()

    :param topic: specifies the desired NSQ topic

    :param channel: specifies the desired NSQ channel

    :param nsqd_tcp_addresses: a sequence of string addresses of the nsqd
        instances this consumer should connect to

    :param lookupd_http_addresses: a sequence of string addresses of the
        nsqlookupd instances this consumer should query for producers of the
        specified topic

    :param name: a string that is used for logging messages (defaults to
        ``'gnsq.consumer.{topic}.{channel}'``)

    :param message_handler: the callable that will be executed for each message
        received

    :param max_tries: the maximum number of attempts the consumer will make to
        process a message after which messages will be automatically discarded

    :param max_in_flight: the maximum number of messages this consumer will
        pipeline for processing. this value will be divided evenly amongst the
        configured/discovered nsqd producers

    :param requeue_delay: the default delay to use when requeueing a failed
        message

    :param lookupd_poll_interval: the amount of time in seconds between querying
        all of the supplied nsqlookupd instances.  A random amount of time based
        on this value will be initially introduced in order to add jitter when
        multiple consumers are running

    :param lookupd_poll_jitter: the maximum fractional amount of jitter to add
        to the lookupd poll loop. This helps evenly distribute requests even if
        multiple consumers restart at the same time.

    :param low_ready_idle_timeout: the amount of time in seconds to wait for a
        message from a producer when in a state where RDY counts are
        re-distributed (ie. `max_in_flight` < `num_producers`)

    :param max_backoff_duration: the maximum time we will allow a backoff state
        to last in seconds. If zero, backoff wil not occur

    :param backoff_on_requeue: if ``False``, backoff will only occur on
        exception

    :param **kwargs: passed to :class:`~gnsq.NsqdTCPClient` initialization
    """
    def __init__(self, topic, channel, nsqd_tcp_addresses=[],
                 lookupd_http_addresses=[], name=None, message_handler=None,
                 max_tries=5, max_in_flight=1, requeue_delay=0,
                 lookupd_poll_interval=60, lookupd_poll_jitter=0.3,
                 low_ready_idle_timeout=10, max_backoff_duration=128,
                 backoff_on_requeue=True, **kwargs):
        if not nsqd_tcp_addresses and not lookupd_http_addresses:
            raise ValueError('must specify at least one nsqd or lookupd')

        self.nsqd_tcp_addresses = parse_nsqds(nsqd_tcp_addresses)
        self.lookupds = parse_lookupds(lookupd_http_addresses)
        self.iterlookupds = cycle(self.lookupds)

        self.topic = topic
        self.channel = channel
        self.max_tries = max_tries
        self.max_in_flight = max_in_flight
        self.requeue_delay = requeue_delay
        self.lookupd_poll_interval = lookupd_poll_interval
        self.lookupd_poll_jitter = lookupd_poll_jitter
        self.low_ready_idle_timeout = low_ready_idle_timeout
        self.backoff_on_requeue = backoff_on_requeue
        self.max_backoff_duration = max_backoff_duration
        self.conn_kwargs = kwargs

        if name:
            self.name = name
        else:
            self.name = '%s.%s.%s' % (__name__, self.topic, self.channel)

        if message_handler is not None:
            self.on_message.connect(message_handler, weak=False)

        self.logger = logging.getLogger(self.name)

        self._state = INIT
        self._redistributed_ready_event = Event()
        self._connection_backoffs = defaultdict(self._create_backoff)
        self._message_backoffs = defaultdict(self._create_backoff)

        self._connections = {}
        self._workers = Group()
        self._killables = Group()

    @cached_property
    def on_message(self):
        """Emitted when a message is received.

        The signal sender is the consumer and the ``message`` is sent as an
        argument. The ``message_handler`` param is connected to this signal.
        """
        return blinker.Signal(doc='Emitted when a message is received.')

    @cached_property
    def on_response(self):
        """Emitted when a response is received.

        The signal sender is the consumer and the ``response`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a response is received.')

    @cached_property
    def on_error(self):
        """Emitted when an error is received.

        The signal sender is the consumer and the ``error`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a error is received.')

    @cached_property
    def on_finish(self):
        """Emitted after a message is successfully finished.

        The signal sender is the consumer and the ``message_id`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted after the a message is finished.')

    @cached_property
    def on_requeue(self):
        """Emitted after a message is requeued.

        The signal sender is the consumer and the ``message_id`` and ``timeout``
        are sent as arguments.
        """
        return blinker.Signal(doc='Emitted after the a message is requeued.')

    @cached_property
    def on_giving_up(self):
        """Emitted after a giving up on a message.

        Emitted when a message has exceeded the maximum number of attempts
        (``max_tries``) and will no longer be requeued. This is useful to
        perform tasks such as writing to disk, collecting statistics etc. The
        signal sender is the consumer and the ``message`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Sent after a giving up on a message.')

    @cached_property
    def on_auth(self):
        """Emitted after a connection is successfully authenticated.

        The signal sender is the consumer and the ``conn`` and parsed
        ``response`` are sent as arguments.
        """
        return blinker.Signal(doc='Emitted when a response is received.')

    @cached_property
    def on_exception(self):
        """Emitted when an exception is caught while handling a message.

        The signal sender is the consumer and the ``message`` and ``error`` are
        sent as arguments.
        """
        return blinker.Signal(doc='Emitted when an exception is caught.')

    @cached_property
    def on_close(self):
        """Emitted after :meth:`close`.

        The signal sender is the consumer.
        """
        return blinker.Signal(doc='Emitted after the consumer is closed.')

    def start(self, block=True):
        """Start discovering and listing to connections."""
        if self._state == INIT:
            if not any(self.on_message.receivers_for(blinker.ANY)):
                raise RuntimeError('no receivers connected to on_message')

            self.logger.debug('starting %s...', self.name)
            self._state = RUNNING
            self.query_nsqd()

            if self.lookupds:
                self.query_lookupd()
                self._killables.add(self._workers.spawn(self._poll_lookupd))

            self._killables.add(self._workers.spawn(self._poll_ready))

        else:
            self.logger.warn('%s already started', self.name)

        if block:
            self.join()

    def close(self):
        """Immediately close all connections and stop workers."""
        if not self.is_running:
            return

        self._state = CLOSED

        self.logger.debug('killing %d worker(s)', len(self._killables))
        self._killables.kill(block=False)

        self.logger.debug('closing %d connection(s)', len(self._connections))
        for conn in self._connections:
            conn.close_stream()

        self.on_close.send(self)

    def join(self, timeout=None, raise_error=False):
        """Block until all connections have closed and workers stopped."""
        self._workers.join(timeout, raise_error)

    @property
    def is_running(self):
        """Check if consumer is currently running."""
        return self._state == RUNNING

    @property
    def is_starved(self):
        """Evaluate whether any of the connections are starved.

        This property should be used by message handlers to reliably identify
        when to process a batch of messages.
        """
        return any(conn.is_starved for conn in self._connections)

    @property
    def total_ready_count(self):
        return sum(c.ready_count for c in self._connections)

    @property
    def total_in_flight(self):
        return sum(c.in_flight for c in self._connections)

    def query_nsqd(self):
        self.logger.debug('querying nsqd...')
        for address in self.nsqd_tcp_addresses:
            address, port = address.split(':')
            self.connect_to_nsqd(address, int(port))

    def query_lookupd(self):
        self.logger.debug('querying lookupd...')
        lookupd = next(self.iterlookupds)

        try:
            producers = lookupd.lookup(self.topic)['producers']
            self.logger.debug('found %d producers', len(producers))

        except Exception as error:
            self.logger.warn(
                'Failed to lookup %s on %s (%s)',
                self.topic, lookupd.address, error)
            return

        for producer in producers:
            self.connect_to_nsqd(
                producer['broadcast_address'], producer['tcp_port'])

    def _poll_lookupd(self):
        try:
            delay = self.lookupd_poll_interval * self.lookupd_poll_jitter
            gevent.sleep(random.random() * delay)

            while True:
                gevent.sleep(self.lookupd_poll_interval)
                self.query_lookupd()

        except gevent.GreenletExit:
            pass

    def _poll_ready(self):
        try:
            while True:
                if self._redistributed_ready_event.wait(5):
                    self._redistributed_ready_event.clear()
                self._redistribute_ready_state()

        except gevent.GreenletExit:
            pass

    def _redistribute_ready_state(self):
        if not self.is_running:
            return

        if len(self._connections) > self.max_in_flight:
            ready_state = self._get_unsaturated_ready_state()
        else:
            ready_state = self._get_saturated_ready_state()

        for conn, count in ready_state.items():
            if conn.ready_count == count:
                self.logger.debug('[%s] RDY count already %d', conn, count)
                continue

            self.logger.debug('[%s] sending RDY %d', conn, count)

            try:
                conn.ready(count)
            except NSQSocketError as error:
                self.logger.warn('[%s] RDY %d failed (%r)', conn, count, error)

    def _get_unsaturated_ready_state(self):
        ready_state = {}
        active = []

        for conn, state in self._connections.items():
            if state == BACKOFF:
                ready_state[conn] = 0

            else:
                active.append(conn)

        random.shuffle(active)

        for conn in active[self.max_in_flight:]:
            ready_state[conn] = 0

        for conn in active[:self.max_in_flight]:
            ready_state[conn] = 1

        return ready_state

    def _get_saturated_ready_state(self):
        ready_state = {}
        active = []
        now = time.time()

        for conn, state in self._connections.items():
            if state == BACKOFF:
                ready_state[conn] = 0

            elif state in (INIT, THROTTLED):
                ready_state[conn] = 1

            elif (now - conn.last_message) > self.low_ready_idle_timeout:
                self.logger.info(
                    '[%s] idle connection, giving up RDY count', conn)
                ready_state[conn] = 1

            else:
                active.append(conn)

        if not active:
            return ready_state

        ready_available = self.max_in_flight - sum(ready_state.values())
        connection_max_in_flight = ready_available // len(active)

        for conn in active:
            ready_state[conn] = connection_max_in_flight

        for conn in random.sample(active, ready_available % len(active)):
            ready_state[conn] += 1

        return ready_state

    def redistribute_ready_state(self):
        self._redistributed_ready_event.set()

    def connect_to_nsqd(self, address, port):
        if not self.is_running:
            return

        conn = NsqdTCPClient(address, port, **self.conn_kwargs)
        if conn in self._connections:
            self.logger.debug('[%s] already connected', conn)
            return

        self._connections[conn] = INIT
        self.logger.debug('[%s] connecting...', conn)

        conn.on_message.connect(self.handle_message)
        conn.on_response.connect(self.handle_response)
        conn.on_error.connect(self.handle_error)
        conn.on_finish.connect(self.handle_finish)
        conn.on_requeue.connect(self.handle_requeue)
        conn.on_auth.connect(self.handle_auth)

        try:
            conn.connect()
            conn.identify()

            if conn.max_ready_count < self.max_in_flight:
                msg = (
                    '[%s] max RDY count %d < consumer max in flight %d, '
                    'truncation possible')

                self.logger.warning(
                    msg, conn, conn.max_ready_count, self.max_in_flight)

            conn.subscribe(self.topic, self.channel)

        except NSQException as error:
            self.logger.warn('[%s] connection failed (%r)', conn, error)
            self.handle_connection_failure(conn)
            return

        # Check if we've closed since we started
        if not self.is_running:
            self.handle_connection_failure(conn)
            return

        self.logger.info('[%s] connection successful', conn)
        self.handle_connection_success(conn)

    def _listen(self, conn):
        try:
            conn.listen()
        except NSQException as error:
            self.logger.warning('[%s] connection lost (%r)', conn, error)

        self.handle_connection_failure(conn)

    def handle_connection_success(self, conn):
        self._workers.spawn(self._listen, conn)
        self.redistribute_ready_state()

        if str(conn) not in self.nsqd_tcp_addresses:
            return

        self._connection_backoffs[conn].success()

    def handle_connection_failure(self, conn):
        del self._connections[conn]
        conn.close_stream()

        if not self.is_running:
            return

        self.redistribute_ready_state()

        if str(conn) not in self.nsqd_tcp_addresses:
            return

        seconds = self._connection_backoffs[conn].failure().get_interval()
        self.logger.debug('[%s] retrying in %ss', conn, seconds)

        gevent.spawn_later(
            seconds, self.connect_to_nsqd, conn.address, conn.port)

    def handle_auth(self, conn, response):
        metadata = []
        if response.get('identity'):
            metadata.append("Identity: %r" % response['identity'])

        if response.get('permission_count'):
            metadata.append("Permissions: %d" % response['permission_count'])

        if response.get('identity_url'):
            metadata.append(response['identity_url'])

        self.logger.info('[%s] AUTH accepted %s', conn, ' '.join(metadata))
        self.on_auth.send(self, conn=conn, response=response)

    def handle_response(self, conn, response):
        self.logger.debug('[%s] response: %s', conn, response)
        self.on_response.send(self, response=response)

    def handle_error(self, conn, error):
        self.logger.debug('[%s] error: %s', conn, error)
        self.on_error.send(self, error=error)

    def _handle_message(self, message):
        if self.max_tries and message.attempts > self.max_tries:
            self.logger.warning(
                "giving up on message '%s' after max tries %d",
                message.id, self.max_tries)
            self.on_giving_up.send(self, message=message)
            return message.finish()

        self.on_message.send(self, message=message)

        if not self.is_running:
            return

        if message.is_async():
            return

        if message.has_responded():
            return

        message.finish()

    def handle_message(self, conn, message):
        self.logger.debug('[%s] got message: %s', conn, message.id)

        try:
            return self._handle_message(message)

        except NSQRequeueMessage as error:
            if error.backoff is None:
                backoff = self.backoff_on_requeue
            else:
                backoff = error.backoff

        except Exception as error:
            backoff = True
            self.logger.exception(
                '[%s] caught exception while handling message', conn)
            self.on_exception.send(self, message=message, error=error)

        if not self.is_running:
            return

        if message.has_responded():
            return

        try:
            message.requeue(self.requeue_delay, backoff)
        except NSQException as error:
            self.logger.warning(
                '[%s] error requeueing message (%r)', conn, error)

    def _create_backoff(self):
        return BackoffTimer(max_interval=self.max_backoff_duration)

    def _start_backoff(self, conn):
        self._connections[conn] = BACKOFF

        interval = self._message_backoffs[conn].get_interval()
        gevent.spawn_later(interval, self._start_throttled, conn)

        self.logger.info('[%s] backing off for %s seconds', conn, interval)
        self.redistribute_ready_state()

    def _start_throttled(self, conn):
        if self._connections.get(conn) != BACKOFF:
            return

        self._connections[conn] = THROTTLED
        self.logger.info('[%s] testing backoff state with RDY 1', conn)
        self.redistribute_ready_state()

    def _complete_backoff(self, conn):
        if self._message_backoffs[conn].is_reset():
            self._connections[conn] = RUNNING
            self.logger.info('backoff complete, resuming normal operation')
            self.redistribute_ready_state()
        else:
            self._start_backoff(conn)

    def _finish_message(self, conn, backoff):
        if not self.max_backoff_duration:
            return

        try:
            state = self._connections[conn]
        except KeyError:
            return

        if state == BACKOFF:
            return

        if backoff:
            self._message_backoffs[conn].failure()
            self._start_backoff(conn)

        elif state == THROTTLED:
            self._message_backoffs[conn].success()
            self._complete_backoff(conn)

        elif state == INIT:
            self._connections[conn] = RUNNING
            self.redistribute_ready_state()

    def handle_finish(self, conn, message_id):
        self.logger.debug('[%s] finished message: %s', conn, message_id)
        self._finish_message(conn, backoff=False)
        self.on_finish.send(self, message_id=message_id)

    def handle_requeue(self, conn, message_id, timeout, backoff):
        self.logger.debug(
            '[%s] requeued message: %s (%s)', conn, message_id, timeout)
        self._finish_message(conn, backoff=backoff)
        self.on_requeue.send(self, message_id=message_id, timeout=timeout)



from __future__ import division, with_statement

import multiprocessing
import os

import pytest
import gevent

from six.moves import range
from gnsq import NsqdHTTPClient, Reader
from gnsq.errors import NSQSocketError

from integration_server import NsqdIntegrationServer


SLOW_TIMEOUT = int(os.environ.get('SLOW_TIMEOUT', '10'), 10)


def test_basic():
    with pytest.raises(ValueError):
        Reader('test', 'test')

    with pytest.raises(TypeError):
        Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=None,
            lookupd_http_addresses='http://localhost:4161/',
        )

    with pytest.raises(TypeError):
        Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses='localhost:4150',
            lookupd_http_addresses=None,
        )

    def message_handler(reader, message):
        pass

    reader = Reader(
        topic='test',
        channel='test',
        name='test',
        max_concurrency=-1,
        nsqd_tcp_addresses='localhost:4150',
        lookupd_http_addresses='http://localhost:4161/',
        message_handler=message_handler
    )

    assert reader.name == 'test'
    assert reader.max_concurrency == multiprocessing.cpu_count()
    assert len(reader.on_message.receivers) == 1

    assert isinstance(reader.nsqd_tcp_addresses, set)
    assert len(reader.nsqd_tcp_addresses) == 1

    assert isinstance(reader.lookupds, list)
    assert len(reader.lookupds) == 1


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_max_concurrency():
    server1 = NsqdIntegrationServer()
    server2 = NsqdIntegrationServer()

    with server1, server2:
        class Accounting(object):
            count = 0
            total = 100
            concurrency = 0
            error = None

        for server in (server1, server2):
            conn = NsqdHTTPClient(server.address, server.http_port)

            for _ in range(Accounting.total // 2):
                conn.publish('test', b'danger zone!')

        reader = Reader(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=[
                server1.tcp_address,
                server2.tcp_address,
            ],
            max_in_flight=5,
            max_concurrency=1,
        )

        @reader.on_exception.connect
        def error_handler(reader, message, error):
            if isinstance(error, NSQSocketError):
                return
            Accounting.error = error
            reader.close()

        @reader.on_message.connect
        def handler(reader, message):
            assert message.body == b'danger zone!'
            assert Accounting.concurrency == 0

            Accounting.concurrency += 1
            gevent.sleep()
            Accounting.concurrency -= 1

            Accounting.count += 1
            if Accounting.count == Accounting.total:
                reader.close()

        reader.start()

        if Accounting.error:
            raise Accounting.error

        assert Accounting.count == Accounting.total



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from mmap import PAGESIZE
from errno import ENOTCONN, EDEADLK, EAGAIN, EWOULDBLOCK

import six

import gevent
from gevent import socket
from gevent.lock import Semaphore
from gevent.ssl import SSLSocket, PROTOCOL_TLSv1_2, CERT_NONE

from gnsq.errors import NSQSocketError

try:
    from .snappy import SnappySocket
except ImportError:
    SnappySocket = None  # pyflakes.ignore

from .defalte import DefalteSocket


class Stream(object):
    def __init__(self, address, port, timeout, buffer_size=PAGESIZE,
                 lock_class=Semaphore):
        self.address = address
        self.port = port
        self.timeout = timeout

        self.buffer = b''
        self.buffer_size = buffer_size

        self.socket = None
        self.lock = lock_class()

    @property
    def is_connected(self):
        return self.socket is not None

    def ensure_connection(self):
        if self.is_connected:
            return
        raise NSQSocketError(ENOTCONN, 'Socket is not connected')

    def connect(self):
        if self.is_connected:
            return

        try:
            self.socket = socket.create_connection(
                address=(self.address, self.port),
                timeout=self.timeout,
            )

        except socket.error as error:
            six.raise_from(NSQSocketError(*error.args), error)

    def read(self, size):
        while len(self.buffer) < size:
            self.ensure_connection()

            try:
                packet = self.socket.recv(self.buffer_size)
            except socket.error as error:
                if error.errno in (EDEADLK, EAGAIN, EWOULDBLOCK):
                    gevent.sleep()
                    continue
                six.raise_from(NSQSocketError(*error.args), error)

            if not packet:
                self.close()

            self.buffer += packet

        data = self.buffer[:size]
        self.buffer = self.buffer[size:]

        return data

    def send(self, data):
        self.ensure_connection()

        with self.lock:
            try:
                return self.socket.sendall(data)
            except socket.error as error:
                six.raise_from(NSQSocketError(*error.args), error)

    def consume_buffer(self):
        data = self.buffer
        self.buffer = b''
        return data

    def close(self):
        if not self.is_connected:
            return

        socket = self.socket
        self.socket = None
        self.buffer = b''

        socket.close()

    def upgrade_to_tls(
        self,
        keyfile=None,
        certfile=None,
        cert_reqs=CERT_NONE,
        ca_certs=None,
        ssl_version=PROTOCOL_TLSv1_2
    ):
        self.ensure_connection()

        try:
            self.socket = SSLSocket(
                self.socket,
                keyfile=keyfile,
                certfile=certfile,
                cert_reqs=cert_reqs,
                ca_certs=ca_certs,
                ssl_version=ssl_version,
            )
        except socket.error as error:
            six.raise_from(NSQSocketError(*error.args), error)

    def upgrade_to_snappy(self):
        if SnappySocket is None:
            raise RuntimeError('snappy requires the python-snappy package')

        self.ensure_connection()
        self.socket = SnappySocket(self.socket)
        self.socket.bootstrap(self.consume_buffer())

    def upgrade_to_defalte(self, level):
        self.ensure_connection()
        self.socket = DefalteSocket(self.socket, level)
        self.socket.bootstrap(self.consume_buffer())



from __future__ import with_statement

import pytest
import gnsq
from integration_server import NsqdIntegrationServer


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_basic():
    with NsqdIntegrationServer() as server:
        conn = gnsq.NsqdHTTPClient(server.address, server.http_port)
        assert conn.ping() == b'OK'
        assert 'topics' in conn.stats()
        assert 'version' in conn.info()


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_topics_channels():
    with NsqdIntegrationServer() as server:
        conn = gnsq.NsqdHTTPClient(server.address, server.http_port)
        assert len(conn.stats()['topics']) == 0

        with pytest.raises(gnsq.errors.NSQHttpError):
            conn.delete_topic('topic')

        conn.create_topic('topic')
        topics = conn.stats()['topics']
        assert len(topics) == 1
        assert topics[0]['topic_name'] == 'topic'

        conn.delete_topic('topic')
        assert len(conn.stats()['topics']) == 0

        with pytest.raises(gnsq.errors.NSQHttpError):
            conn.create_channel('topic', 'channel')

        with pytest.raises(gnsq.errors.NSQHttpError):
            conn.delete_channel('topic', 'channel')

        conn.create_topic('topic')
        assert len(conn.stats()['topics'][0]['channels']) == 0

        conn.create_channel('topic', 'channel')
        channels = conn.stats()['topics'][0]['channels']
        assert len(channels) == 1
        assert channels[0]['channel_name'] == 'channel'

        conn.delete_channel('topic', 'channel')
        assert len(conn.stats()['topics'][0]['channels']) == 0


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_publish():
    with NsqdIntegrationServer() as server:
        conn = gnsq.NsqdHTTPClient(server.address, server.http_port)

        conn.publish('topic', b'sup')
        assert conn.stats()['topics'][0]['depth'] == 1

        conn.multipublish('topic', [b'sup', b'sup'])
        assert conn.stats()['topics'][0]['depth'] == 3

        conn.multipublish('topic', iter([b'sup', b'sup', b'sup']))
        assert conn.stats()['topics'][0]['depth'] == 6

        conn.empty_topic('topic')
        assert conn.stats()['topics'][0]['depth'] == 0

        conn.create_topic('topic')
        conn.create_channel('topic', 'channel')
        conn.publish('topic', b'sup')
        assert conn.stats()['topics'][0]['channels'][0]['depth'] == 1

        conn.empty_channel('topic', 'channel')
        assert conn.stats()['topics'][0]['channels'][0]['depth'] == 0

        if server.version < (0, 3, 6):
            return

        conn.publish('topic', b'sup', 60 * 1000)
        stats = conn.stats()
        assert stats['topics'][0]['channels'][0]['depth'] == 0
        assert stats['topics'][0]['channels'][0]['deferred_count'] == 1



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json

import urllib3

from .errors import NSQHttpError
from .version import __version__

USERAGENT = 'gnsq/{}'.format(__version__)


def _encode(value):
    if isinstance(value, bytes):
        return value
    return value.encode('utf-8')


def _encode_dict(value):
    if value is None:
        return None
    return {_encode(k): _encode(v) for k, v in value.items()}


class HTTPClient(object):
    @classmethod
    def from_url(cls, url, **kwargs):
        """Create a client from a url."""
        url = urllib3.util.parse_url(url)
        if url.host:
            kwargs.setdefault('host', url.host)

        if url.port:
            kwargs.setdefault('port', url.port)

        if url.scheme == 'https':
            kwargs.setdefault('connection_class', urllib3.HTTPSConnectionPool)

        return cls(**kwargs)

    def __init__(self, host, port, useragent=USERAGENT,
                 connection_class=urllib3.HTTPConnectionPool, **kwargs):
        self.useragent = useragent
        self._connection = connection_class(host, port, **kwargs)

    @property
    def scheme(self):
        return self._connection.scheme

    @property
    def host(self):
        return self._connection.host

    @property
    def port(self):
        return self._connection.port

    @property
    def address(self):
        return '{}://{}:{}/'.format(self.scheme, self.host, self.port)

    def _request(self, method, url, headers={}, fields=None, **kwargs):
        headers = dict(headers)
        headers.setdefault('Accept', 'application/vnd.nsq version=1.0')
        headers.setdefault('User-Agent', self.useragent)

        response = self._connection.request_encode_url(
            method, url, headers=_encode_dict(headers),
            fields=_encode_dict(fields), **kwargs)

        if 'application/json' in response.getheader('content-type', ''):
            return self._http_check_json(response)

        return self._http_check(response)

    def _http_check(self, response):
        if response.status != 200:
            raise NSQHttpError('http error <{}>'.format(response.status))
        return response.data

    def _http_check_json(self, response):
        try:
            data = json.loads(response.data.decode('utf-8'))
        except ValueError:
            return self._http_check(response)

        if response.status != 200:
            status_txt = data.get('status_txt', 'http error')
            raise NSQHttpError('{} <{}>'.format(status_txt, response.status))

        # Handle 1.0.0-compat vs 0.x versions
        try:
            data['data']
        except KeyError:
            return data

    def __repr__(self):
        return '<{!s} {!r}>'.format(type(self).__name__, self.address)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import blinker
from .decorators import cached_property
from .errors import NSQException


class Message(object):
    """A class representing a message received from nsqd."""
    def __init__(self, timestamp, attempts, id, body):
        self.timestamp = timestamp
        self.attempts = attempts
        self.id = id
        self.body = body
        self._has_responded = False
        self._is_async = False

    @cached_property
    def on_finish(self):
        """Emitted after :meth:`finish`.

        The signal sender is the message instance.
        """
        return blinker.Signal(doc='Emitted after message is finished.')

    @cached_property
    def on_requeue(self):
        """Emitted after :meth:`requeue`.

        The signal sender is the message instance and sends the ``timeout`` and
        a ``backoff`` flag as arguments.
        """
        return blinker.Signal(doc='Emitted after message is requeued.')

    @cached_property
    def on_touch(self):
        """Emitted after :meth:`touch`.

        The signal sender is the message instance.
        """
        return blinker.Signal(doc='Emitted after message is touched.')

    def enable_async(self):
        """Enables asynchronous processing for this message.

        :class:`~gnsq.Consumer` will not automatically respond to the message
        upon return of :meth:`~gnsq.Consumer.handle_message`.
        """
        self._is_async = True

    def is_async(self):
        """Returns whether or not asynchronous processing has been enabled."""
        return self._is_async

    def has_responded(self):
        """Returns whether or not this message has been responded to."""
        return self._has_responded

    def finish(self):
        """
        Respond to nsqd that you’ve processed this message successfully
        (or would like to silently discard it).
        """
        if self._has_responded:
            raise NSQException('already responded')
        self._has_responded = True
        self.on_finish.send(self)

    def requeue(self, time_ms=0, backoff=True):
        """
        Respond to nsqd that you’ve failed to process this message successfully
        (and would like it to be requeued).
        """
        if self._has_responded:
            raise NSQException('already responded')
        self._has_responded = True
        self.on_requeue.send(self, timeout=time_ms, backoff=backoff)

    def touch(self):
        """Respond to nsqd that you need more time to process the message."""
        if self._has_responded:
            raise NSQException('already responded')
        self.on_touch.send(self)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division

import random
from multiprocessing import cpu_count

from gevent.queue import Queue

from .consumer import Consumer
from .decorators import deprecated
from .errors import NSQNoConnections
from .states import INIT


class Reader(Consumer):
    """Use :class:`~gnsq.Consumer` instead.

    .. deprecated:: 1.0.0
    """
    @deprecated
    def __init__(self, *args, **kwargs):
        """Use :class:`~gnsq.Consumer` instead.

        .. deprecated:: 1.0.0
        """
        setattr(self, 'async', kwargs.pop('async', False))

        max_concurrency = kwargs.pop('max_concurrency', 0)

        if max_concurrency < 0:
            self.max_concurrency = cpu_count()
        else:
            self.max_concurrency = max_concurrency

        if self.max_concurrency:
            self.queue = Queue()
        else:
            self.queue = None

        super(Reader, self).__init__(*args, **kwargs)

    def start(self, *args, **kwargs):
        if self._state == INIT:
            for _ in range(self.max_concurrency):
                self._killables.add(self._workers.spawn(self._run))

        return super(Reader, self).start(*args, **kwargs)

    def handle_message(self, conn, message):
        if self.max_concurrency:
            self.logger.debug('[%s] queueing message: %s' % (conn, message.id))
            self.queue.put((conn, message))
        else:
            super(Reader, self).handle_message(conn, message)

    @deprecated
    def publish(self, topic, message):
        """Use :class:`~gnsq.Producer` instead.

        .. deprecated:: 1.0.0
        """
        if not self.connections:
            raise NSQNoConnections()
        conn = random.choice(list(self.connections))
        conn.publish(topic, message)

    def _handle_message(self, message):
        if getattr(self, 'async'):
            message.enable_async()
        return super(Reader, self)._handle_message(message)

    def _run(self):
        for conn, message in self.queue:
            if not conn.is_connected:
                continue
            super(Reader, self).handle_message(conn, message)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import urllib3

from . import protocol as nsq
from .decorators import deprecated
from .httpclient import HTTPClient


class LookupdClient(HTTPClient):
    """Low level http client for nsqlookupd.

    :param host: nsqlookupd host address (default: localhost)

    :param port: nsqlookupd http port (default: 4161)

    :param useragent: useragent sent to nsqlookupd (default:
        ``<client_library_name>/<version>``)

    :param connection_class: override the http connection class
    """
    def __init__(self, host='localhost', port=4161, **kwargs):
        super(LookupdClient, self).__init__(host, port, **kwargs)

    def lookup(self, topic):
        """Returns producers for a topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('GET', '/lookup', fields={'topic': topic})

    def topics(self):
        """Returns all known topics."""
        return self._request('GET', '/topics')

    def channels(self, topic):
        """Returns all known channels of a topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('GET', '/channels', fields={'topic': topic})

    def nodes(self):
        """Returns all known nsqd."""
        return self._request('GET', '/nodes')

    def create_topic(self, topic):
        """Add a topic to nsqlookupd's registry."""
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/create', fields={'topic': topic})

    def delete_topic(self, topic):
        """Deletes an existing topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/delete', fields={'topic': topic})

    def create_channel(self, topic, channel):
        """Add a channel to nsqlookupd's registry."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self._request('POST', '/channel/create',
                             fields={'topic': topic, 'channel': channel})

    def delete_channel(self, topic, channel):
        """Deletes an existing channel of an existing topic."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self._request('POST', '/channel/delete',
                             fields={'topic': topic, 'channel': channel})

    def tombstone_topic(self, topic, node):
        """Tombstones a specific producer of an existing topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/tombstone',
                             fields={'topic': topic, 'node': node})

    def ping(self):
        """Monitoring endpoint.

        :returns: should return `"OK"`, otherwise raises an exception.
        """
        return self._request('GET', '/ping')

    def info(self):
        """Returns version information."""
        return self._request('GET', '/info')


class Lookupd(LookupdClient):
    """Use :class:`LookupdClient` instead.

    .. deprecated:: 1.0.0
    """

    @deprecated
    def __init__(self, address='http://localhost:4161/', **kwargs):
        """Use :meth:`LookupdClient.from_url` instead.

        .. deprecated:: 1.0.0
        """
        self.address = self.base_url = address

        url = urllib3.util.parse_url(address)
        if url.host:
            kwargs.setdefault('host', url.host)

        if url.port:
            kwargs.setdefault('port', url.port)

        if url.scheme == 'https':
            kwargs.setdefault('connection_class', urllib3.HTTPSConnectionPool)

        return super(Lookupd, self).__init__(**kwargs)

    @property
    @deprecated
    def base_url(self):
        """Use :attr:`LookupdClient.address` instead.

        .. deprecated:: 1.0.0
        """
        return self.address

    @deprecated
    def tombstone_topic_producer(self, topic, node):
        """Use :meth:`LookupdClient.tombstone_topic` instead.

        .. deprecated:: 1.0.0
        """
        return self.tombstone_topic(topic, node)



# -*- coding: utf-8 -*-
import itertools
import json
import sys

import gnsq


class LogGiveupHandler(object):
    """Log messages on giveup.

    Writes the message body to the log. This can be customized by subclassing
    and implementing :meth:`format_message`. Assuming messages do not requeued
    using the `to_nsq` utility.

    Example usage::

        >>> fp = open('topic.__BURY__.log', 'w')
        >>> consumer.on_giving_up.connect(
        ...     LogGiveupHandler(fp.write), weak=False)
    """
    def __init__(self, log=sys.stdout.write, newline='\n'):
        self.log = log
        self.newline = newline

    def format_message(self, message):
        return message.body

    def __call__(self, consumer, message):
        self.log(self.format_message(message) + self.newline)


class JSONLogGiveupHandler(LogGiveupHandler):
    """Log messages as json on giveup.

    Works like :class:`LogGiveupHandler` but serializes the message details as
    json before writing to the log.

    Example usage::

        >>> fp = open('topic.__BURY__.log', 'w')
        >>> consumer.on_giving_up.connect(
        ...     JSONLogGiveupHandler(fp.write), weak=False)
    """
    def format_message(self, message):
        return json.dumps({
            'timestamp': message.timestamp,
            'attempts': message.attempts,
            'id': message.id,
            'body': message.body,
        })


class NsqdGiveupHandler(object):
    """Send messages by to nsq on giveup.

    Forwards the message body to the given topic where it can be inspected and
    requeued. This can be customized by subclassing and implementing
    :meth:`format_message`. Messages can be requeued with the `nsq_to_nsq`
    utility.

    Example usage::

        >>> giveup_handler = NsqdGiveupHandler('topic.__BURY__')
        >>> consumer.on_giving_up.connect(giveup_handler)
    """

    def __init__(self, topic, nsqd_hosts=['localhost'],
                 nsqd_class=gnsq.NsqdHTTPClient):
        if not nsqd_hosts:
            raise ValueError('at least one nsqd host is required')
        self.topic = topic
        self.nsqds = itertools.cycle([nsqd_class(host) for host in nsqd_hosts])

    def format_message(self, message):
        return message.body

    def __call__(self, consumer, message):
        nsq = self.nsqds.next()
        nsq.publish(self.topic, self.format_message(message))



import errno
import io
import os
import random
import re
import shutil
import subprocess
import sys
import tempfile


class BaseIntegrationServer(object):
    protocols = ('TCP', 'HTTP')

    protocol_re = re.compile(' '.join([
        r'(?P<protocol>[A-Z]+):',
        r'listening on',
        r'(?P<address>(?:[0-9]{1,3}\.){3}[0-9]{1,3}):(?P<port>[0-9]+)',
    ]))

    version_re = re.compile(' '.join([
        r'(?P<name>[a-z]+)',
        r'v(?P<version>[0-9]+\.[0-9]+\.[0-9]+)(?:-[a-z]+)?',
        r'\(built w\/(?P<go_version>[a-z0-9.-]+)\)'
    ]))

    def __init__(self, address='127.0.0.1'):
        self.address = address
        self.protocol_ports = {}
        self.data_path = tempfile.mkdtemp()
        self.name, self.version, self.go_version = self._parse_version()

    def _parse_version(self):
        output = subprocess.check_output([self.executable, '--version'])
        match = self.version_re.match(output.decode('utf-8'))
        name = match.group('name')
        version = tuple(int(v) for v in match.group('version').split('.'))
        go_version = match.group('go_version')
        return name, version, go_version

    @property
    def tcp_port(self):
        return self.protocol_ports['TCP']

    @property
    def tcp_address(self):
        return '%s:%d' % (self.address, self.tcp_port)

    @property
    def http_port(self):
        return self.protocol_ports['HTTP']

    @property
    def http_address(self):
        return '%s:%d' % (self.address, self.http_port)

    def _random_port(self):
        if self.version < (0, 3, 5):
            return random.randint(10000, 65535)
        return 0

    def _random_address(self):
        return '%s:%d' % (self.address, self._random_port())

    def _parse_protocol_ports(self):
        while len(self.protocol_ports) < len(self.protocols):
            line = self.child.stderr.readline()
            sys.stderr.write(line)

            if not line:
                raise Exception('server exited prematurely')

            if 'listening on' not in line:
                continue

            match = self.protocol_re.search(line)
            if not match:
                raise Exception('unexpected line: %r' % line)

            protocol = match.group('protocol')
            if protocol not in self.protocols:
                continue

            port = int(match.group('port'), 10)
            self.protocol_ports[protocol] = port

    def __enter__(self):
        sys.stderr.write('running: %s\n' % ' '.join(self.cmd))
        self.child = subprocess.Popen(self.cmd, stderr=subprocess.PIPE)
        if sys.version_info[0] == 3:
            self.child.stderr = io.TextIOWrapper(self.child.stderr, 'utf-8')
        self._parse_protocol_ports()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.child.terminate()
        except OSError as error:
            if error.errno == errno.ESRCH:
                return
            raise

        while True:
            line = self.child.stderr.readline()
            if not line:
                break
            sys.stderr.write(line)

        self.child.wait()
        shutil.rmtree(self.data_path)


class NsqdIntegrationServer(BaseIntegrationServer):
    executable = 'nsqd'

    tls_cert = os.path.join(os.path.dirname(__file__), 'cert.pem')
    tls_key = os.path.join(os.path.dirname(__file__), 'key.pem')

    def __init__(self, lookupd=None, **kwargs):
        super(NsqdIntegrationServer, self).__init__(**kwargs)

        if self.has_https():
            self.protocols = ('TCP', 'HTTP', 'HTTPS')

        self.lookupd = lookupd

    def has_https(self):
        return self.version >= (0, 2, 28)

    @property
    def https_port(self):
        return self.protocol_ports['HTTPS']

    @property
    def https_address(self):
        return '%s:%d' % (self.address, self.https_port)

    @property
    def cmd(self):
        cmd = [
            self.executable,
            '--broadcast-address', self.address,
            '--tcp-address', self._random_address(),
            '--http-address', self._random_address(),
            '--data-path', self.data_path,
            '--tls-cert', self.tls_cert,
            '--tls-key', self.tls_key,
        ]

        if self.has_https():
            cmd.extend(['--https-address', self._random_address()])

        if self.lookupd:
            cmd.extend(['--lookupd-tcp-address', self.lookupd])

        return cmd


class LookupdIntegrationServer(BaseIntegrationServer):
    executable = 'nsqlookupd'

    @property
    def cmd(self):
        return [
            self.executable,
            '--broadcast-address', self.address,
            '--tcp-address', self._random_address(),
            '--http-address', self._random_address(),
        ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json
import time

import blinker

from gevent import socket

from . import protocol as nsq
from . import errors

from .decorators import cached_property, deprecated
from .httpclient import HTTPClient, USERAGENT
from .message import Message
from .states import CONNECTED, DISCONNECTED, INIT
from .stream import Stream

HOSTNAME = socket.gethostname()
SHORTNAME = HOSTNAME.split('.')[0]


class NsqdTCPClient(object):
    """Low level object representing a TCP connection to nsqd.

    :param address: the host or ip address of the nsqd

    :param port: the nsqd tcp port to connect to

    :param timeout: the timeout for read/write operations (in seconds)

    :param client_id: an identifier used to disambiguate this client (defaults
        to the first part of the hostname)

    :param hostname: the hostname where the client is deployed (defaults to the
        clients hostname)

    :param heartbeat_interval: the amount of time in seconds to negotiate with
        the connected producers to send heartbeats (requires nsqd 0.2.19+)

    :param output_buffer_size: size of the buffer (in bytes) used by nsqd for
        buffering writes to this connection

    :param output_buffer_timeout: timeout (in ms) used by nsqd before flushing
        buffered writes (set to 0 to disable). Warning: configuring clients with
        an extremely low (< 25ms) output_buffer_timeout has a significant effect
        on nsqd CPU usage (particularly with > 50 clients connected).

    :param tls_v1: enable TLS v1 encryption (requires nsqd 0.2.22+)

    :param tls_options: dictionary of options to pass to `ssl.wrap_socket()
        <http://docs.python.org/2/library/ssl.html#ssl.wrap_socket>`_

    :param snappy: enable Snappy stream compression (requires nsqd 0.2.23+)

    :param deflate: enable deflate stream compression (requires nsqd 0.2.23+)

    :param deflate_level: configure the deflate compression level for this
        connection (requires nsqd 0.2.23+)

    :param sample_rate: take only a sample of the messages being sent to the
        client. Not setting this or setting it to 0 will ensure you get all the
        messages destined for the client. Sample rate can be greater than 0 or
        less than 100 and the client will receive that percentage of the message
        traffic. (requires nsqd 0.2.25+)

    :param auth_secret: a string passed when using nsq auth (requires
        nsqd 0.2.29+)

    :param user_agent: a string identifying the agent for this client in the
        spirit of HTTP (default: ``<client_library_name>/<version>``) (requires
        nsqd 0.2.25+)
    """
    def __init__(
        self,
        address='127.0.0.1',
        port=4150,
        timeout=60.0,
        client_id=None,
        hostname=None,
        heartbeat_interval=30,
        output_buffer_size=16 * 1024,
        output_buffer_timeout=250,
        tls_v1=False,
        tls_options=None,
        snappy=False,
        deflate=False,
        deflate_level=6,
        sample_rate=0,
        auth_secret=None,
        user_agent=USERAGENT,
    ):
        self.address = address
        self.port = port
        self.timeout = timeout

        self.client_id = client_id or SHORTNAME
        self.hostname = hostname or HOSTNAME
        self.heartbeat_interval = 1000 * heartbeat_interval
        self.output_buffer_size = output_buffer_size
        self.output_buffer_timeout = output_buffer_timeout
        self.tls_v1 = tls_v1
        self.tls_options = tls_options
        self.snappy = snappy
        self.deflate = deflate
        self.deflate_level = deflate_level
        self.sample_rate = sample_rate
        self.auth_secret = auth_secret
        self.user_agent = user_agent

        self.state = INIT
        self.last_response = time.time()
        self.last_message = time.time()
        self.ready_count = 0
        self.in_flight = 0
        self.max_ready_count = 2500

        self._frame_handlers = {
            nsq.FRAME_TYPE_RESPONSE: self.handle_response,
            nsq.FRAME_TYPE_ERROR: self.handle_error,
            nsq.FRAME_TYPE_MESSAGE: self.handle_message
        }

    @cached_property
    def on_message(self):
        """Emitted when a message frame is received.

        The signal sender is the connection and the ``message`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a message frame is received.')

    @cached_property
    def on_response(self):
        """Emitted when a response frame is received.

        The signal sender is the connection and the ``response`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a response frame is received.')

    @cached_property
    def on_error(self):
        """Emitted when an error frame is received.

        The signal sender is the connection and the ``error`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a error frame is received.')

    @cached_property
    def on_finish(self):
        """Emitted after :meth:`finish`.

        Sent after a message owned by this connection is successfully finished.
        The signal sender is the connection and the ``message_id`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted after the a message is finished.')

    @cached_property
    def on_requeue(self):
        """Emitted after :meth:`requeue`.

        Sent after a message owned by this connection is requeued. The signal
        sender is the connection and the ``message_id``, ``timeout`` and
        ``backoff`` flag are sent as arguments.
        """
        return blinker.Signal(doc='Emitted after the a message is requeued.')

    @cached_property
    def on_auth(self):
        """Emitted after the connection is successfully authenticated.

        The signal sender is the connection and the parsed ``response`` is sent
        as arguments.
        """
        return blinker.Signal(
            doc='Emitted after the connection is successfully authenticated.')

    @cached_property
    def on_close(self):
        """Emitted after :meth:`close_stream`.

        Sent after the connection socket has closed. The signal sender is the
        connection.
        """
        return blinker.Signal(doc='Emitted after the connection is closed.')

    @property
    def is_connected(self):
        """Check if the client is currently connected."""
        return self.state == CONNECTED

    @property
    def is_starved(self):
        """Evaluate whether the connection is starved.

        This property should be used by message handlers to reliably identify
        when to process a batch of messages.
        """
        return self.in_flight >= max(self.ready_count * 0.85, 1)

    def connect(self):
        """Initialize connection to the nsqd."""
        if self.state == DISCONNECTED:
            raise errors.NSQException('connection already closed')

        if self.is_connected:
            return

        stream = Stream(self.address, self.port, self.timeout)
        stream.connect()

        self.stream = stream
        self.state = CONNECTED
        self.send(nsq.MAGIC_V2)

    def close_stream(self):
        """Close the underlying socket."""
        if not self.is_connected:
            return

        self.stream.close()
        self.state = DISCONNECTED
        self.on_close.send(self)

    def send(self, data):
        try:
            return self.stream.send(data)
        except Exception:
            self.close_stream()
            raise

    def _read_response(self):
        try:
            size = nsq.unpack_size(self.stream.read(4))
            return self.stream.read(size)
        except Exception:
            self.close_stream()
            raise

    def read_response(self):
        """Read an individual response from nsqd.

        :returns: tuple of the frame type and the processed data.
        """
        response = self._read_response()
        frame, data = nsq.unpack_response(response)
        self.last_response = time.time()

        if frame not in self._frame_handlers:
            raise errors.NSQFrameError('unknown frame {}'.format(frame))

        frame_handler = self._frame_handlers[frame]
        processed_data = frame_handler(data)

        return frame, processed_data

    def handle_response(self, data):
        if data == nsq.HEARTBEAT:
            self.nop()

        self.on_response.send(self, response=data)
        return data

    def handle_error(self, data):
        error = errors.make_error(data)
        self.on_error.send(self, error=error)

        if error.fatal:
            self.close_stream()

        return error

    def handle_message(self, data):
        self.last_message = time.time()
        self.in_flight += 1

        message = Message(*nsq.unpack_message(data))
        message.on_finish.connect(self.handle_finish)
        message.on_requeue.connect(self.handle_requeue)
        message.on_touch.connect(self.handle_touch)

        self.on_message.send(self, message=message)
        return message

    def handle_finish(self, message):
        self.finish(message.id)

    def handle_requeue(self, message, timeout, backoff):
        self.requeue(message.id, timeout, backoff)

    def handle_touch(self, message):
        self.touch(message.id)

    def finish_inflight(self):
        self.in_flight -= 1

    def listen(self):
        """Listen to incoming responses until the connection closes."""
        while self.is_connected:
            self.read_response()

    def check_ok(self, expected=nsq.OK):
        frame, data = self.read_response()
        if frame == nsq.FRAME_TYPE_ERROR:
            raise data

        if frame != nsq.FRAME_TYPE_RESPONSE:
            raise errors.NSQException('expected response frame')

        if data != expected:
            raise errors.NSQException('unexpected response {!r}'.format(data))

    def upgrade_to_tls(self):
        self.stream.upgrade_to_tls(**self.tls_options)
        self.check_ok()

    def upgrade_to_snappy(self):
        self.stream.upgrade_to_snappy()
        self.check_ok()

    def upgrade_to_defalte(self):
        self.stream.upgrade_to_defalte(self.deflate_level)
        self.check_ok()

    def identify(self):
        """Update client metadata on the server and negotiate features.

        :returns: nsqd response data if there was feature negotiation,
            otherwise ``None``
        """
        self.send(nsq.identify({
            # nsqd 0.2.28+
            'client_id': self.client_id,
            'hostname': self.hostname,

            # nsqd 0.2.19+
            'feature_negotiation': True,
            'heartbeat_interval': self.heartbeat_interval,

            # nsqd 0.2.21+
            'output_buffer_size': self.output_buffer_size,
            'output_buffer_timeout': self.output_buffer_timeout,

            # nsqd 0.2.22+
            'tls_v1': self.tls_v1,

            # nsqd 0.2.23+
            'snappy': self.snappy,
            'deflate': self.deflate,
            'deflate_level': self.deflate_level,

            # nsqd nsqd 0.2.25+
            'sample_rate': self.sample_rate,
            'user_agent': self.user_agent,
        }))

        frame, data = self.read_response()

        if frame == nsq.FRAME_TYPE_ERROR:
            raise data

        if data == nsq.OK:
            return

        try:
            data = json.loads(data.decode('utf-8'))

        except ValueError:
            self.close_stream()
            raise errors.NSQException(
                'failed to parse IDENTIFY response JSON from nsqd: '
                '{!r}'.format(data))

        self.max_ready_count = data.get('max_rdy_count', self.max_ready_count)

        if self.tls_v1 and data.get('tls_v1'):
            self.upgrade_to_tls()

        if self.snappy and data.get('snappy'):
            self.upgrade_to_snappy()

        elif self.deflate and data.get('deflate'):
            self.deflate_level = data.get('deflate_level', self.deflate_level)
            self.upgrade_to_defalte()

        if self.auth_secret and data.get('auth_required'):
            self.auth()

        return data

    def auth(self):
        """Send authorization secret to nsqd."""
        self.send(nsq.auth(self.auth_secret))
        frame, data = self.read_response()

        if frame == nsq.FRAME_TYPE_ERROR:
            raise data

        try:
            response = json.loads(data.decode('utf-8'))
        except ValueError:
            self.close_stream()
            raise errors.NSQException(
                'failed to parse AUTH response JSON from nsqd: '
                '{!r}'.format(data))

        self.on_auth.send(self, response=response)
        return response

    def subscribe(self, topic, channel):
        """Subscribe to a nsq `topic` and `channel`."""
        self.send(nsq.subscribe(topic, channel))

    def publish(self, topic, data, defer=None):
        """Publish a message to the given topic over tcp.

        :param topic: the topic to publish to

        :param data: bytestring data to publish

        :param defer: duration in milliseconds to defer before publishing
            (requires nsq 0.3.6)
        """
        if defer is None:
            self.send(nsq.publish(topic, data))
        else:
            self.send(nsq.deferpublish(topic, data, defer))

    def multipublish(self, topic, messages):
        """Publish an iterable of messages to the given topic over http.

        :param topic: the topic to publish to

        :param messages: iterable of bytestrings to publish
        """
        self.send(nsq.multipublish(topic, messages))

    def ready(self, count):
        """Indicate you are ready to receive ``count`` messages."""
        self.ready_count = count
        self.send(nsq.ready(count))

    def finish(self, message_id):
        """Finish a message (indicate successful processing)."""
        self.send(nsq.finish(message_id))
        self.finish_inflight()
        self.on_finish.send(self, message_id=message_id)

    def requeue(self, message_id, timeout=0, backoff=True):
        """Re-queue a message (indicate failure to process)."""
        self.send(nsq.requeue(message_id, timeout))
        self.finish_inflight()
        self.on_requeue.send(
            self,
            message_id=message_id,
            timeout=timeout,
            backoff=backoff
        )

    def touch(self, message_id):
        """Reset the timeout for an in-flight message."""
        self.send(nsq.touch(message_id))

    def close(self):
        """Indicate no more messages should be sent."""
        self.send(nsq.close())

    def nop(self):
        """Send no-op to nsqd. Used to keep connection alive."""
        self.send(nsq.nop())

    def __str__(self):
        return '{}:{}'.format(self.address, self.port)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return isinstance(other, type(self)) and str(self) == str(other)

    def __cmp__(self, other):
        return hash(self) - hash(other)

    def __lt__(self, other):
        return hash(self) < hash(other)


class NsqdHTTPClient(HTTPClient):
    """Low level http client for nsqd.

    :param host: nsqd host address (default: localhost)

    :param port: nsqd http port (default: 4151)

    :param useragent: useragent sent to nsqd (default:
        ``<client_library_name>/<version>``)

    :param connection_class: override the http connection class
    """
    def __init__(self, host='localhost', port=4151, **kwargs):
        super(NsqdHTTPClient, self).__init__(host, port, **kwargs)

    def publish(self, topic, data, defer=None):
        """Publish a message to the given topic over http.

        :param topic: the topic to publish to

        :param data: bytestring data to publish

        :param defer: duration in millisconds to defer before publishing
            (requires nsq 0.3.6)
        """
        nsq.assert_valid_topic_name(topic)
        fields = {'topic': topic}

        if defer is not None:
            fields['defer'] = '{}'.format(defer)

        return self._request('POST', '/pub', fields=fields, body=data)

    def _validate_mpub_message(self, message):
        if b'\n' not in message:
            return message
        raise errors.NSQException(
            'newlines are not allowed in http multipublish')

    def multipublish(self, topic, messages, binary=False):
        """Publish an iterable of messages to the given topic over http.

        :param topic: the topic to publish to

        :param messages: iterable of bytestrings to publish

        :param binary: enable binary mode. defaults to False
            (requires nsq 1.0.0)

        By default multipublish expects messages to be delimited by ``"\\n"``,
        use the binary flag to enable binary mode where the POST body is
        expected to be in the following wire protocol format.
        """
        nsq.assert_valid_topic_name(topic)
        fields = {'topic': topic}

        if binary:
            fields['binary'] = 'true'
            body = nsq.multipublish_body(messages)
        else:
            body = b'\n'.join(self._validate_mpub_message(m) for m in messages)

        return self._request('POST', '/mpub', fields=fields, body=body)

    def create_topic(self, topic):
        """Create a topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/create', fields={'topic': topic})

    def delete_topic(self, topic):
        """Delete a topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/delete', fields={'topic': topic})

    def create_channel(self, topic, channel):
        """Create a channel for an existing topic."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self._request('POST', '/channel/create',
                             fields={'topic': topic, 'channel': channel})

    def delete_channel(self, topic, channel):
        """Delete an existing channel for an existing topic."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self._request('POST', '/channel/delete',
                             fields={'topic': topic, 'channel': channel})

    def empty_topic(self, topic):
        """Empty all the queued messages for an existing topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/empty', fields={'topic': topic})

    def empty_channel(self, topic, channel):
        """Empty all the queued messages for an existing channel."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self._request('POST', '/channel/empty',
                             fields={'topic': topic, 'channel': channel})

    def pause_topic(self, topic):
        """Pause message flow to all channels on an existing topic.

        Messages will queue at topic.
        """
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/pause', fields={'topic': topic})

    def unpause_topic(self, topic):
        """Resume message flow to channels of an existing, paused, topic."""
        nsq.assert_valid_topic_name(topic)
        return self._request('POST', '/topic/unpause', fields={'topic': topic})

    def pause_channel(self, topic, channel):
        """Pause message flow to consumers of an existing channel.

        Messages will queue at channel.
        """
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self._request('POST', '/channel/pause',
                             fields={'topic': topic, 'channel': channel})

    def unpause_channel(self, topic, channel):
        """Resume message flow to consumers of an existing, paused, channel."""
        nsq.assert_valid_topic_name(topic)
        nsq.assert_valid_channel_name(channel)
        return self._request('POST', '/channel/unpause',
                             fields={'topic': topic, 'channel': channel})

    def stats(self, topic=None, channel=None, text=False):
        """Return internal instrumented statistics.

        :param topic: (optional) filter to topic

        :param channel: (optional) filter to channel

        :param text: return the stats as a string (default: ``False``)
        """
        if text:
            fields = {'format': 'text'}
        else:
            fields = {'format': 'json'}

        if topic:
            nsq.assert_valid_topic_name(topic)
            fields['topic'] = topic

        if channel:
            nsq.assert_valid_channel_name(channel)
            fields['channel'] = channel

        return self._request('GET', '/stats', fields=fields)

    def ping(self):
        """Monitoring endpoint.

        :returns: should return ``"OK"``, otherwise raises an exception.
        """
        return self._request('GET', '/ping')

    def info(self):
        """Returns version information."""
        return self._request('GET', '/info')


class Nsqd(object):
    """Use :class:`NsqdTCPClient` or :class:`NsqdHTTPClient` instead.

    .. deprecated:: 1.0.0
    """
    @deprecated
    def __init__(self, address='127.0.0.1', tcp_port=4150, http_port=4151,
                 **kwargs):
        """Use :class:`NsqdTCPClient` or :class:`NsqdHTTPClient` instead.

        .. deprecated:: 1.0.0
        """
        self.address = address
        self.tcp_port = tcp_port
        self.http_port = http_port

        self.__tcp_client = NsqdTCPClient(address, tcp_port, **kwargs)
        self.__http_client = NsqdHTTPClient(address, http_port)

    @property
    def base_url(self):
        return 'http://{}:{}/'.format(self.address, self.http_port)

    @deprecated
    def publish_tcp(self, topic, data, **kwargs):
        """Use :meth:`NsqdTCPClient.publish` instead.

        .. deprecated:: 1.0.0
        """
        return self.__tcp_client.publish(topic, data, **kwargs)

    @deprecated
    def publish_http(self, topic, data, **kwargs):
        """Use :meth:`NsqdHTTPClient.publish` instead.

        .. deprecated:: 1.0.0
        """
        self.__http_client.publish(topic, data, **kwargs)

    def publish(self, topic, data, *args, **kwargs):
        if self.__tcp_client.is_connected:
            return self.__tcp_client.publish(topic, data, *args, **kwargs)
        else:
            return self.__http_client.publish(topic, data, *args, **kwargs)

    @deprecated
    def multipublish_tcp(self, topic, messages, **kwargs):
        """Use :meth:`NsqdTCPClient.multipublish` instead.

        .. deprecated:: 1.0.0
        """
        return self.__tcp_client.multipublish(topic, messages, **kwargs)

    @deprecated
    def multipublish_http(self, topic, messages, **kwargs):
        """Use :meth:`NsqdHTTPClient.multipublish` instead.

        .. deprecated:: 1.0.0
        """
        return self.__http_client.multipublish(topic, messages, **kwargs)

    def multipublish(self, topic, messages, *args, **kwargs):
        if self.__tcp_client.is_connected:
            return self.__tcp_client.multipublish(
                topic, messages, *args, **kwargs)
        else:
            return self.__http_client.multipublish(
                topic, messages, *args, **kwargs)

    def __getattr__(self, name):
        for client in (self.__tcp_client, self.__http_client):
            try:
                return getattr(client, name)
            except AttributeError:
                pass

        return super(Nsqd, self).__getattr__(name)

    def __str__(self):
        return '{}:{}'.format(self.address, self.tcp_port)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return isinstance(other, type(self)) and str(self) == str(other)

    def __cmp__(self, other):
        return hash(self) - hash(other)

    def __lt__(self, other):
        return hash(self) < hash(other)



# -*- coding: utf-8 -*-
import gevent.queue


class QueueHandler(gevent.queue.Queue):
    """Iterator like api for gnsq.

    Example usage:
    >>> queue = QueueHandler()
    >>> consumer = Consumer('topic', 'worker', max_in_flight=16)
    >>> consumer.on_message.connect(queue)
    >>> consumer.start(block=False)
    >>> for message in queue:
    ...     print(message.body)
    ...     message.finish()

    Or give it to a pool:
    >>> gevent.pool.Pool().map(queue, my_handler)
    """
    def __init__(self, heartbeat=5, timeout=None, spawn=None):
        self.timeout = timeout
        self.heartbeat = heartbeat

    def __call__(self, consumer, message):
        message.enable_async()
        self.put(message)



from __future__ import division, with_statement

import os

import pytest
import gevent

from gnsq import NsqdHTTPClient, Producer
from gnsq.errors import NSQException, NSQNoConnections, NSQInvalid

from integration_server import NsqdIntegrationServer


SLOW_TIMEOUT = int(os.environ.get('SLOW_TIMEOUT', '10'), 10)


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_publish():
    with NsqdIntegrationServer() as server:
        producer = Producer(server.tcp_address)
        producer.start()

        for _ in range(100):
            producer.publish('test', b'hi')

        producer.close()
        producer.join()

        conn = NsqdHTTPClient(server.address, server.http_port)
        stats = conn.stats()

        assert stats['topics'][0]['depth'] == 100


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_async_publish():
    with NsqdIntegrationServer() as server:
        results = []
        producer = Producer(server.tcp_address)
        producer.start()

        for _ in range(100):
            results.append(producer.publish('test', b'hi', raise_error=False))

        gevent.joinall(results, raise_error=True)
        producer.close()
        producer.join()

        conn = NsqdHTTPClient(server.address, server.http_port)
        stats = conn.stats()

        assert stats['topics'][0]['depth'] == 100


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_multipublish():
    with NsqdIntegrationServer() as server:
        producer = Producer(server.tcp_address)
        producer.start()

        for _ in range(10):
            producer.multipublish('test', 10 * [b'hi'])

        producer.close()
        producer.join()

        conn = NsqdHTTPClient(server.address, server.http_port)
        stats = conn.stats()

        assert stats['topics'][0]['depth'] == 100


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_async_multipublish():
    with NsqdIntegrationServer() as server:
        results = []
        producer = Producer(server.tcp_address)
        producer.start()

        for _ in range(10):
            result = producer.multipublish(
                'test', 10 * [b'hi'], raise_error=False)
            results.append(result)

        gevent.joinall(results, raise_error=True)
        producer.close()
        producer.join()

        conn = NsqdHTTPClient(server.address, server.http_port)
        stats = conn.stats()

        assert stats['topics'][0]['depth'] == 100


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_publish_error():
    with NsqdIntegrationServer() as server:
        producer = Producer(server.tcp_address)
        producer.start()

        with pytest.raises(NSQInvalid):
            producer.publish('test', b'hi', defer=-1000)

        producer.close()
        producer.join()


def test_not_running():
    producer = Producer('192.0.2.1:4150')

    with pytest.raises(NSQException):
        producer.publish('topic', b'hi')


def test_no_connections():
    producer = Producer('192.0.2.1:4150', timeout=0.01)
    producer.start()

    with pytest.raises(NSQNoConnections):
        producer.publish('topic', b'hi', block=False)



import pytest
import gnsq


class MockConnection(object):
    def __init__(self, message, operations):
        message.on_finish.connect(self.finish)
        message.on_requeue.connect(self.requeue)
        message.on_touch.connect(self.touch)
        self.operations = iter(operations)

    def finish(self, message):
        exp_name, exp_args = next(self.operations)
        assert exp_name == 'finish'
        assert exp_args == (message,)

    def requeue(self, message, timeout, backoff):
        exp_name, exp_args = next(self.operations)
        assert exp_name == 'requeue'
        assert exp_args == (message, timeout, backoff)

    def touch(self, message):
        exp_name, exp_args = next(self.operations)
        assert exp_name == 'touch'
        assert exp_args == (message,)

    def assert_finished(self):
        with pytest.raises(StopIteration):
            next(self.operations)


def test_basic():
    message = gnsq.Message(0, 42, '1234', 'sup')
    assert message.timestamp == 0
    assert message.attempts == 42
    assert message.id == '1234'
    assert message.body == 'sup'
    assert message.has_responded() is False


def test_finish():
    message = gnsq.Message(0, 42, '1234', 'sup')
    mock_conn = MockConnection(message, [
        ('finish', (message,)),
    ])
    assert message.has_responded() is False

    message.finish()
    assert message.has_responded() is True

    with pytest.raises(gnsq.errors.NSQException):
        message.finish()

    mock_conn.assert_finished()


def test_requeue():
    message = gnsq.Message(0, 42, '1234', 'sup')
    mock_conn = MockConnection(message, [
        ('requeue', (message, 0, True)),
    ])
    assert message.has_responded() is False

    message.requeue()
    assert message.has_responded() is True

    with pytest.raises(gnsq.errors.NSQException):
        message.requeue()

    mock_conn.assert_finished()


def test_requeue_timeout():
    message = gnsq.Message(0, 42, '1234', 'sup')
    mock_conn = MockConnection(message, [
        ('requeue', (message, 1000, True)),
    ])
    assert message.has_responded() is False

    message.requeue(1000)
    assert message.has_responded() is True

    with pytest.raises(gnsq.errors.NSQException):
        message.requeue(1000)

    mock_conn.assert_finished()


def test_backoff():
    message = gnsq.Message(0, 42, '1234', 'sup')
    mock_conn = MockConnection(message, [
        ('requeue', (message, 0, False)),
    ])
    assert message.has_responded() is False

    message.requeue(backoff=False)
    assert message.has_responded() is True

    with pytest.raises(gnsq.errors.NSQException):
        message.requeue()

    mock_conn.assert_finished()


def test_touch():
    message = gnsq.Message(0, 42, '1234', 'sup')
    mock_conn = MockConnection(message, [
        ('touch', (message,)),
        ('touch', (message,)),
        ('touch', (message,)),
        ('finish', (message,)),
    ])
    assert message.has_responded() is False

    message.touch()
    message.touch()
    message.touch()
    assert message.has_responded() is False

    message.finish()
    assert message.has_responded() is True

    with pytest.raises(gnsq.errors.NSQException):
        message.touch()

    mock_conn.assert_finished()



import json
import struct

import pytest
import six

from gnsq import protocol as nsq


IDENTIFY_DICT_ASCII = {'a': 1, 'b': 2}
IDENTIFY_DICT_UNICODE = {'c': u'w\xc3\xa5\xe2\x80\xa0'}
IDENTIFY_BODY_ASCII = six.b(json.dumps(IDENTIFY_DICT_ASCII))
IDENTIFY_BODY_UNICODE = six.b(json.dumps(IDENTIFY_DICT_UNICODE))

MSGS = [b'asdf', b'ghjk', b'abcd']
MPUB_BODY = struct.pack('>l', len(MSGS)) + b''.join(struct.pack('>l', len(m)) + m for m in MSGS)


@pytest.mark.parametrize('cmd_method,kwargs,result', [
    ('identify',
        {'data': IDENTIFY_DICT_ASCII},
        b'IDENTIFY\n' + struct.pack('>l', len(IDENTIFY_BODY_ASCII)) +
        IDENTIFY_BODY_ASCII),
    ('identify',
        {'data': IDENTIFY_DICT_UNICODE},
        b'IDENTIFY\n' + struct.pack('>l', len(IDENTIFY_BODY_UNICODE)) +
        IDENTIFY_BODY_UNICODE),
    ('auth',
        {'secret': b'secret'},
        b'AUTH\n' + struct.pack('>l', 6) + b'secret'),
    ('subscribe',
        {'topic_name': 'test_topic', 'channel_name': 'test_channel'},
        b'SUB test_topic test_channel\n'),
    ('finish',
        {'message_id': 'test'},
        b'FIN test\n'),
    ('finish',
        {'message_id': u'\u2020est \xfcn\xee\xe7\xf8\u2202\xe9'},
        b'FIN \xe2\x80\xa0est \xc3\xbcn\xc3\xae\xc3\xa7\xc3\xb8\xe2\x88\x82\xc3\xa9\n'),
    ('requeue',
        {'message_id': 'test'},
        b'REQ test 0\n'),
    ('requeue',
        {'message_id': 'test', 'timeout': 60},
        b'REQ test 60\n'),
    ('touch',
        {'message_id': 'test'},
        b'TOUCH test\n'),
    ('ready',
        {'count': 100},
        b'RDY 100\n'),
    ('nop',
        {},
        b'NOP\n'),
    ('publish',
        {'topic_name': 'test', 'data': MSGS[0]},
        b'PUB test\n' + struct.pack('>l', len(MSGS[0])) + MSGS[0]),
    ('multipublish',
        {'topic_name': 'test', 'messages': MSGS},
        b'MPUB test\n' + struct.pack('>l', len(MPUB_BODY)) + MPUB_BODY),
    ('deferpublish',
        {'topic_name': 'test', 'data': MSGS[0], 'delay_ms': 42},
        b'DPUB test 42\n' + struct.pack('>l', len(MSGS[0])) + MSGS[0]),
])
def test_command(cmd_method, kwargs, result):
    assert getattr(nsq, cmd_method)(**kwargs) == result


def test_unicode_body():
    pytest.raises(TypeError, nsq.publish, 'topic', u'unicode body')



import pytest

from six.moves import range
from gnsq import BackoffTimer
from gnsq import protocol as nsq


@pytest.mark.parametrize('name,good', [
    ('valid_name', True),
    ('invalid name with space', False),
    ('invalid_name_due_to_length_this_is_' + (4 * 'really_') + 'long', False),
    ('test-with_period.', True),
    ('test#ephemeral', True),
    ('test:ephemeral', False),
])
def test_topic_names(name, good):
    assert nsq.valid_topic_name(name) == good


@pytest.mark.parametrize('name,good', [
    ('test', True),
    ('test-with_period.', True),
    ('test#ephemeral', True),
    ('invalid_name_due_to_length_this_is_' + (4 * 'really_') + 'long', False),
    ('invalid name with space', False),
])
def test_channel_names(name, good):
    assert nsq.valid_channel_name(name) == good


def test_assert_topic():
    assert nsq.assert_valid_topic_name('topic') is None

    with pytest.raises(ValueError):
        nsq.assert_valid_topic_name('invalid name with space')


def test_assert_channel():
    assert nsq.assert_valid_channel_name('channel') is None

    with pytest.raises(ValueError):
        nsq.assert_valid_channel_name('invalid name with space')


def test_invalid_commands():
    with pytest.raises(TypeError):
        nsq.requeue('1234', None)

    with pytest.raises(TypeError):
        nsq.ready(None)

    with pytest.raises(ValueError):
        nsq.ready(-1)


def test_backoff_timer():
    timer = BackoffTimer(max_interval=1000)
    assert timer.get_interval() == 0
    assert timer.is_reset()

    timer.success()
    assert timer.get_interval() == 0
    assert timer.is_reset()

    timer.failure()
    assert timer.c == 1
    assert not timer.is_reset()

    for _ in range(100):
        interval = timer.get_interval()
        assert interval > 0 and interval < 2

    timer.failure()
    assert timer.c == 2

    for _ in range(100):
        interval = timer.get_interval()
        assert interval > 0 and interval < 4

    timer.success().success()
    assert timer.get_interval() == 0
    assert timer.is_reset()

    for _ in range(100):
        timer.failure()

    assert timer.c == 100
    assert timer.get_interval() <= 1000

    timer.reset()
    assert timer.c == 0
    assert timer.get_interval() == 0

    timer = BackoffTimer(min_interval=1000)
    assert timer.c == 0
    assert timer.get_interval() == 1000



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division

import logging
from collections import defaultdict, deque

import blinker
import gevent

from gevent.event import AsyncResult
from gevent.pool import Group
from gevent.queue import Queue, Empty

from . import protocol as nsq

from .backofftimer import BackoffTimer
from .decorators import cached_property
from .errors import NSQException, NSQNoConnections
from .nsqd import NsqdTCPClient
from .states import INIT, RUNNING, CLOSED
from .util import parse_nsqds


class Producer(object):
    """High level NSQ producer.

    A Producer will connect to the nsqd tcp addresses and support async
    publishing (``PUB`` & ``MPUB`` & ``DPUB``) of messages to `nsqd` over the
    TCP protocol.

    Example publishing a message::

        from gnsq import Producer

        producer = Producer('localhost:4150')
        producer.start()
        producer.publish('topic', b'hello world')

    :param nsqd_tcp_addresses: a sequence of string addresses of the nsqd
        instances this consumer should connect to

    :param max_backoff_duration: the maximum time we will allow a backoff state
        to last in seconds. If zero, backoff wil not occur

    :param **kwargs: passed to :class:`~gnsq.NsqdTCPClient` initialization
    """
    def __init__(self, nsqd_tcp_addresses=[], max_backoff_duration=128,
                 **kwargs):
        if not nsqd_tcp_addresses:
            raise ValueError('must specify at least one nsqd or lookupd')

        self.nsqd_tcp_addresses = parse_nsqds(nsqd_tcp_addresses)
        self.max_backoff_duration = max_backoff_duration
        self.conn_kwargs = kwargs
        self.logger = logging.getLogger(__name__)

        self._state = INIT
        self._connections = Queue()
        self._connection_backoffs = defaultdict(self._create_backoff)
        self._response_queues = {}
        self._workers = Group()

    @cached_property
    def on_response(self):
        """Emitted when a response is received.

        The signal sender is the consumer and the ` ` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a response is received.')

    @cached_property
    def on_error(self):
        """Emitted when an error is received.

        The signal sender is the consumer and the ``error`` is sent as an
        argument.
        """
        return blinker.Signal(doc='Emitted when a error is received.')

    @cached_property
    def on_auth(self):
        """Emitted after a connection is successfully authenticated.

        The signal sender is the consumer and the ``conn`` and parsed
        ``response`` are sent as arguments.
        """
        return blinker.Signal(doc='Emitted when a response is received.')

    @cached_property
    def on_close(self):
        """Emitted after :meth:`close`.

        The signal sender is the consumer.
        """
        return blinker.Signal(doc='Emitted after the consumer is closed.')

    def start(self):
        """Start discovering and listing to connections."""
        if self._state == CLOSED:
            raise NSQException('producer already closed')

        if self.is_running:
            self.logger.warn('producer already started')
            return

        self.logger.debug('starting producer...')
        self._state = RUNNING

        for address in self.nsqd_tcp_addresses:
            address, port = address.split(':')
            self.connect_to_nsqd(address, int(port))

    def close(self):
        """Immediately close all connections and stop workers."""
        if not self.is_running:
            return

        self._state = CLOSED
        self.logger.debug('closing connection(s)')

        while True:
            try:
                conn = self._connections.get(block=False)
            except Empty:
                break

            conn.close_stream()

        self.on_close.send(self)

    def join(self, timeout=None, raise_error=False):
        """Block until all connections have closed and workers stopped."""
        self._workers.join(timeout, raise_error)

    @property
    def is_running(self):
        """Check if the producer is currently running."""
        return self._state == RUNNING

    def connect_to_nsqd(self, address, port):
        if not self.is_running:
            return

        conn = NsqdTCPClient(address, port, **self.conn_kwargs)
        self.logger.debug('[%s] connecting...', conn)

        conn.on_response.connect(self.handle_response)
        conn.on_error.connect(self.handle_error)
        conn.on_auth.connect(self.handle_auth)

        try:
            conn.connect()
            conn.identify()

        except NSQException as error:
            self.logger.warn('[%s] connection failed (%r)', conn, error)
            self.handle_connection_failure(conn)
            return

        # Check if we've closed since we started
        if not self.is_running:
            self.handle_connection_failure(conn)
            return

        self.logger.info('[%s] connection successful', conn)
        self.handle_connection_success(conn)

    def _listen(self, conn):
        try:
            conn.listen()
        except NSQException as error:
            self.logger.warning('[%s] connection lost (%r)', conn, error)

        self.handle_connection_failure(conn)

    def handle_connection_success(self, conn):
        self._response_queues[conn] = deque()
        self._put_connection(conn)
        self._workers.spawn(self._listen, conn)
        self._connection_backoffs[conn].success()

    def handle_connection_failure(self, conn):
        conn.close_stream()
        self._clear_responses(conn, NSQException('connection closed'))

        if not self.is_running:
            return

        seconds = self._connection_backoffs[conn].failure().get_interval()
        self.logger.debug('[%s] retrying in %ss', conn, seconds)

        gevent.spawn_later(
            seconds, self.connect_to_nsqd, conn.address, conn.port)

    def handle_auth(self, conn, response):
        metadata = []
        if response.get('identity'):
            metadata.append("Identity: %r" % response['identity'])

        if response.get('permission_count'):
            metadata.append("Permissions: %d" % response['permission_count'])

        if response.get('identity_url'):
            metadata.append(response['identity_url'])

        self.logger.info('[%s] AUTH accepted %s', conn, ' '.join(metadata))
        self.on_auth.send(self, conn=conn, response=response)

    def handle_response(self, conn, response):
        self.logger.debug('[%s] response: %s', conn, response)

        if response == nsq.OK:
            result = self._response_queues[conn].popleft()
            result.set(response)

        self.on_response.send(self, response=response)

    def handle_error(self, conn, error):
        self.logger.debug('[%s] error: %s', conn, error)
        self._clear_responses(conn, error)
        self.on_error.send(self, error=error)

    def _create_backoff(self):
        return BackoffTimer(max_interval=self.max_backoff_duration)

    def _clear_responses(self, conn, error):
        # All relevent errors are fatal
        for result in self._response_queues.pop(conn, []):
            result.set_exception(error)

    def _get_connection(self, block=True, timeout=None):
        if not self.is_running:
            raise NSQException('producer not running')

        while True:
            try:
                conn = self._connections.get(block=block, timeout=timeout)
            except Empty:
                raise NSQNoConnections

            if conn.is_connected:
                return conn

            # Discard closed connections

    def _put_connection(self, conn):
        if not self.is_running:
            return
        self._connections.put(conn)

    def publish(self, topic, data, defer=None, block=True, timeout=None,
                raise_error=True):
        """Publish a message to the given topic.

        :param topic: the topic to publish to

        :param data: bytestring data to publish

        :param defer: duration in milliseconds to defer before publishing
            (requires nsq 0.3.6)

        :param block: wait for a connection to become available before
            publishing the message. If block is `False` and no connections
            are available, :class:`~gnsq.errors.NSQNoConnections` is raised

        :param timeout: if timeout is a positive number, it blocks at most
            ``timeout`` seconds before raising
            :class:`~gnsq.errors.NSQNoConnections`

        :param raise_error: if ``True``, it blocks until a response is received
            from the nsqd server, and any error response is raised. Otherwise
            an :class:`~gevent.event.AsyncResult` is returned
        """
        result = AsyncResult()
        conn = self._get_connection(block=block, timeout=timeout)

        try:
            self._response_queues[conn].append(result)
            conn.publish(topic, data, defer=defer)
        finally:
            self._put_connection(conn)

        if raise_error:
            return result.get()

        return result

    def multipublish(self, topic, messages, block=True, timeout=None,
                     raise_error=True):
        """Publish an iterable of messages to the given topic.

        :param topic: the topic to publish to

        :param messages: iterable of bytestrings to publish

        :param block: wait for a connection to become available before
            publishing the message. If block is `False` and no connections
            are available, :class:`~gnsq.errors.NSQNoConnections` is raised

        :param timeout: if timeout is a positive number, it blocks at most
            ``timeout`` seconds before raising
            :class:`~gnsq.errors.NSQNoConnections`

        :param raise_error: if ``True``, it blocks until a response is received
            from the nsqd server, and any error response is raised. Otherwise
            an :class:`~gevent.event.AsyncResult` is returned
        """
        result = AsyncResult()
        conn = self._get_connection(block=block, timeout=timeout)

        try:
            self._response_queues[conn].append(result)
            conn.multipublish(topic, messages)
        finally:
            self._put_connection(conn)

        if raise_error:
            return result.get()

        return result



# -*- coding: utf-8 -*-


class SentryExceptionHandler(object):
    """Log gnsq exceptions to sentry.

    Example usage::

        >>> from raven import Sentry
        >>> sentry = Sentry()
        >>> consumer.on_exception.connect(
        ...     SentryExceptionHandler(sentry), weak=False)
    """

    def __init__(self, client):
        self.client = client

    def message_extra(self, message):
        return {
            'id': message.id,
            'timestamp': message.timestamp,
            'attempts': message.attempts,
            'body': message.body,
        }

    def __call__(self, consumer, message, error):
        extra = {}

        if message:
            extra['message'] = self.message_extra(message)

        self.client.captureException(extra=extra)



# -*- coding: utf-8 -*-
import logging
import warnings

import gevent.queue
import gevent.pool

from gnsq.errors import NSQException


TIMEOUT_WARNING = 'batching timed out. batch size may be to large'
STARVED_WARNING = 'consumer is starved. batch size may be to large'


class BatchHandler(object):
    """Batch message handler for gnsq.

    It is recommended to use a max inflight greater than the batch size.

    Example usage::

        >>> consumer = Consumer('topic', 'worker', max_in_flight=16)
        >>> consumer.on_message.connect(BatchHandler(8, my_handler), weak=False)
    """
    def __init__(self, batch_size, handle_batch=None, handle_message=None,
                 handle_batch_error=None, handle_message_error=None,
                 timeout=10, spawn=gevent.spawn):
        self.logger = logging.getLogger(__name__)
        self.message_channel = gevent.queue.Channel()
        self.batch_size = batch_size
        self.timeout = timeout

        if isinstance(spawn, int):
            spawn = gevent.pool.Pool(spawn).spawn

        self.spawn = spawn

        if handle_batch is not None:
            self.handle_batch = handle_batch

        if handle_message is not None:
            self.handle_message = handle_message

        if handle_batch_error is not None:
            self.handle_batch_error = handle_batch_error

        if handle_message_error is not None:
            self.handle_message_error = handle_message_error

        self.worker = gevent.spawn(self._run)

    def __call__(self, consumer, message):
        message.enable_async()
        self.message_channel.put(message)

        if consumer.is_starved:
            self.message_channel.put(StopIteration)

    def _run(self):
        while True:
            messages = []

            while len(messages) < self.batch_size:
                try:
                    message = self.message_channel.get(timeout=self.timeout)
                except gevent.queue.Empty:
                    warnings.warn(TIMEOUT_WARNING, RuntimeWarning)
                    break

                if message is StopIteration:
                    warnings.warn(STARVED_WARNING, RuntimeWarning)
                    break

                messages.append(message)

            if messages:
                self.spawn(self.run_batch, messages)

    def finish_message(self, message):
        try:
            message.finish()
        except NSQException as error:
            self.logger.warning('error finishing message (%r)', error)

    def finish_messages(self, messages):
        for message in messages:
            if message.has_responded():
                continue
            self.finish_message(message)

    def requeue_message(self, message):
        try:
            message.requeue()
        except NSQException as error:
            self.logger.warning('error requeueing message (%r)', error)

    def requeue_messages(self, messages):
        for message in messages:
            if message.has_responded():
                continue
            self.requeue_message(message)

    def run_batch(self, messages):
        batch = []

        for message in messages:
            try:
                batch.append(self.handle_message(message))
            except Exception as error:
                self.logger.exception('caught exception while handling message')
                self.handle_message_error(error, message)
                self.requeue_message(message)

        if batch:
            try:
                self.handle_batch(batch)
            except Exception as error:
                self.logger.exception('caught exception while handling batch')
                self.handle_batch_error(error, messages, batch)
                self.requeue_messages(messages)
                return

        self.finish_messages(messages)

    def handle_message(self, message):
        """Handle a single message.

        Over ride this to provide some processing and an individual message.
        The result of this function is what is passed to :meth:`handle_batch`.
        This may be overridden or passed into the constructor. By default it
        simply returns the message.

        Raising an exception in :meth:`handle_message` will cause that message
        to be requeued and excluded from the batch.
        """
        return message

    def handle_batch(self, messages):
        """Handle a batch message.

        Processes a batch of messages. You must provide a :meth:`handle_batch`
        function to the constructor or override this method.

        Raising an exception in :meth:`handle_batch` will cause all messages in
        the batch to be requeued.
        """
        raise RuntimeError('handle_message must be overridden')

    def handle_message_error(self, error, message):
        """Handle an exception processesing an individual message.

        This may be overridden or passed into the constructor.
        """
        pass

    def handle_batch_error(self, error, messages, batch):
        """Handle an exception processsing a batch of messages.

        This may be overridden or passed into the constructor.
        """
        pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import snappy
from .compression import CompressionSocket


class SnappySocket(CompressionSocket):
    def __init__(self, socket):
        self._decompressor = snappy.StreamDecompressor()
        self._compressor = snappy.StreamCompressor()
        super(SnappySocket, self).__init__(socket)

    def compress(self, data):
        return self._compressor.add_chunk(data, compress=True)

    def decompress(self, data):
        return self._decompressor.decompress(data)



import random

import six

from .lookupd import LookupdClient


def normalize_nsqd_address(address):
    if not isinstance(address, six.string_types):
        raise TypeError('nsqd address must be a string')

    host, _, port = address.partition(':')
    if port:
        try:
            port = int(port, 10)
        except ValueError:
            raise ValueError('invalid nsqd port')

    else:
        port = 4150

    return '{host}:{port}'.format(host=host, port=port)


def parse_nsqds(nsqd_tcp_addresses):
    if isinstance(nsqd_tcp_addresses, six.string_types):
        return set([normalize_nsqd_address(nsqd_tcp_addresses)])

    elif isinstance(nsqd_tcp_addresses, (list, tuple, set)):
        return set(normalize_nsqd_address(addr) for addr in nsqd_tcp_addresses)

    raise TypeError('nsqd_tcp_addresses must be a list, set or tuple')


def parse_lookupds(lookupd_http_addresses):
    if isinstance(lookupd_http_addresses, six.string_types):
        return [LookupdClient.from_url(lookupd_http_addresses)]

    if not isinstance(lookupd_http_addresses, (list, tuple)):
        msg = 'lookupd_http_addresses must be a list, set or tuple'
        raise TypeError(msg)

    lookupd = [LookupdClient.from_url(a) for a in lookupd_http_addresses]
    random.shuffle(lookupd)

    return lookupd



from gevent.event import AsyncResult
from gevent.server import StreamServer


class mock_server(object):
    def __init__(self, handler):
        self.handler = handler
        self.result = AsyncResult()
        self.server = StreamServer(('127.0.0.1', 0), self)

    def __call__(self, socket, address):
        try:
            self.result.set(self.handler(socket, address))
        except Exception as error:
            self.result.set_exception(error)
        finally:
            socket.close()

    def __enter__(self):
        self.server.start()
        return self.server

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.result.get()
        self.server.stop()



from __future__ import division, with_statement

import os

import pytest
import gevent

from gnsq import NsqdHTTPClient, Consumer, states
from gnsq.errors import NSQSocketError

from integration_server import LookupdIntegrationServer, NsqdIntegrationServer


SLOW_TIMEOUT = int(os.environ.get('SLOW_TIMEOUT', '10'), 10)


def test_basic():
    with pytest.raises(ValueError):
        Consumer('test', 'test')

    with pytest.raises(TypeError):
        Consumer(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=None,
            lookupd_http_addresses='http://localhost:4161/',
        )

    with pytest.raises(TypeError):
        Consumer(
            topic='test',
            channel='test',
            nsqd_tcp_addresses='localhost:4150',
            lookupd_http_addresses=None,
        )

    def message_handler(consumer, message):
        pass

    consumer = Consumer(
        topic='test',
        channel='test',
        name='test',
        nsqd_tcp_addresses='localhost:4150',
        lookupd_http_addresses='http://localhost:4161/',
        message_handler=message_handler
    )

    assert consumer.name == 'test'
    assert len(consumer.on_message.receivers) == 1

    assert isinstance(consumer.nsqd_tcp_addresses, set)
    assert len(consumer.nsqd_tcp_addresses) == 1

    assert isinstance(consumer.lookupds, list)
    assert len(consumer.lookupds) == 1


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_messages():
    with NsqdIntegrationServer() as server:

        class Accounting(object):
            count = 0
            total = 500
            error = None

        conn = NsqdHTTPClient(server.address, server.http_port)
        for _ in range(Accounting.total):
            conn.publish('test', b'danger zone!')

        consumer = Consumer(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=[server.tcp_address],
            max_in_flight=100,
        )

        @consumer.on_exception.connect
        def error_handler(consumer, message, error):
            if isinstance(error, NSQSocketError):
                return
            Accounting.error = error
            consumer.close()

        @consumer.on_message.connect
        def handler(consumer, message):
            assert message.body == b'danger zone!'

            Accounting.count += 1
            if Accounting.count == Accounting.total:
                consumer.close()

        consumer.start()

        if Accounting.error:
            raise Accounting.error

        assert Accounting.count == Accounting.total


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_lookupd():
    with LookupdIntegrationServer() as lookupd_server:
        server1 = NsqdIntegrationServer(lookupd=lookupd_server.tcp_address)
        server2 = NsqdIntegrationServer(lookupd=lookupd_server.tcp_address)

        with server1, server2:
            class Accounting(object):
                count = 0
                total = 500
                concurrency = 0
                error = None

            for server in (server1, server2):
                conn = NsqdHTTPClient(server.address, server.http_port)

                for _ in range(Accounting.total // 2):
                    conn.publish('test', b'danger zone!')

            consumer = Consumer(
                topic='test',
                channel='test',
                lookupd_http_addresses=lookupd_server.http_address,
                max_in_flight=32,
            )

            @consumer.on_exception.connect
            def error_handler(consumer, message, error):
                if isinstance(error, NSQSocketError):
                    return
                Accounting.error = error
                consumer.close()

            @consumer.on_message.connect
            def handler(consumer, message):
                assert message.body == b'danger zone!'

                Accounting.count += 1
                if Accounting.count == Accounting.total:
                    consumer.close()

            gevent.sleep(0.1)
            consumer.start()

            if Accounting.error:
                raise Accounting.error

            assert Accounting.count == Accounting.total


@pytest.mark.slow
@pytest.mark.timeout(SLOW_TIMEOUT)
def test_backoff():
    with NsqdIntegrationServer() as server:
        conn = NsqdHTTPClient(server.address, server.http_port)

        for _ in range(500):
            conn.publish('test', 'danger zone!')

        consumer = Consumer(
            topic='test',
            channel='test',
            nsqd_tcp_addresses=[server.tcp_address],
            max_in_flight=100,
            message_handler=lambda consumer, message: None
        )

        consumer.start(block=False)
        consumer._redistributed_ready_event.wait()

        conn = next(iter(consumer._connections))
        consumer._message_backoffs[conn].failure()
        consumer._message_backoffs[conn].failure()
        consumer._start_backoff(conn)
        consumer._redistribute_ready_state()

        assert consumer._connections[conn] == states.BACKOFF
        assert consumer.total_ready_count == 0

        consumer._start_throttled(conn)
        consumer._redistribute_ready_state()
        consumer._redistribute_ready_state()

        assert consumer._connections[conn] == states.THROTTLED
        assert consumer.total_ready_count == 1

        consumer._message_backoffs[conn].success()
        consumer._complete_backoff(conn)
        consumer._redistribute_ready_state()

        assert consumer._connections[conn] == states.BACKOFF
        assert consumer.total_ready_count == 0

        consumer._start_throttled(conn)
        consumer._redistribute_ready_state()

        assert consumer._connections[conn] == states.THROTTLED
        assert consumer.total_ready_count == 1

        consumer._message_backoffs[conn].success()
        consumer._complete_backoff(conn)
        consumer._redistribute_ready_state()

        assert consumer._connections[conn] == states.RUNNING
        assert consumer.total_ready_count == 100


def test_no_handlers():
    consumer = Consumer('test', 'test', 'localhost:4150')
    with pytest.raises(RuntimeError):
        consumer.start(block=False)
